/*-----------------------------------------------------------------------------
 *	楕円面(Obj)と光線(視点E,方向余弦ベクトルV)の方程式の解Tを求める
 *-----------------------------------------------------------------------------*/

#include <math.h>
#include "ellipse.h"
#include "vector.h"
#include "matrix.h"
#include "color.h"
#include "plight.h"

double calcEllipseT(const Ellipse Obj, const Vector E, const Vector V){
	double a,b,c;	/*係数、定数項*/
	double D;	/* 判別式 */
	Matrix M;
	Vector R0;

	M = IdentityMatrix;

	/* コードを書こう */
	M = ScaleMatrix(1.0/(Obj.a*Obj.a),1.0/(Obj.b*Obj.b),1.0/(Obj.c*Obj.c));
	R0 = sub(E,Obj.pos);

	a = dot(tVectorMatrix(V,M),V);
	b = dot(tVectorMatrix(V,M),R0);
	c = dot(tVectorMatrix(R0,M),R0)-1; 
	D = b*b - a*c;

	if(D<0){
		return(-1);	/*解なし*/
	}else if (D==0){
		return(-b/a);
	}else {
		return((-b-sqrt(D))/a);
	}
}

/*-----------------------------------------------------------------------------
 * 楕円面(Obj)上の点Pにおける単位法線ベクトルを求める
 *---------------------------------------------------------------------------*/
Vector calcEllipseNorm(const Ellipse Obj, const Vector P){
	Vector N;
	Matrix M;

	M = ScaleMatrix(1/(Obj.a*Obj.a),1/(Obj.b*Obj.b),1/(Obj.c*Obj.c));
	N = tVectorMatrix(sub(P,Obj.pos), M);
	N = normalize(N);

	return(N);
}

/*----------------------------------------------------------------------*/
/* 陰影計算を行う							*/
/* src:対象となる楕円 	Q:光源						*/
/* P:交点 		N:交点における法線 	E:視線位置		*/
/* ia:環境光強度 	id:拡散光強度		is:鏡面反射光強度	*/
/*----------------------------------------------------------------------*/
void calcShading(const Ellipse src, const PointLight Q,
		 const Vector P, const Vector N, const Vector E,
		 double *ia, double *id, double *is){

	Vector R;	/* 反射光ベクトル(R) */
	Vector L;	/* 光源ベクトル(L) */
	Vector V;	/* 視線ベクトル(V) */
	double r;	/* 交点、光源位置との距離 */
	double dot_N_L;	/* ２つのベクトルNとLとの内積を求める作業変数 */
 	double dot_V_R;	/* ２つのベクトルRとVとの内積を求める作業変数 */
	
	/* 環境光の強度 */
	*ia = src.ka;

	/* 交点と光源との距離を求める */
	L = sub(Q.pos, P);
	r = length(L);
	L = normalize(L);

	/* dot_N_L */
	dot_N_L = dot(N, L);
	/* 陰か影の中かの処理 */
	if (dot_N_L<0){      /* 陰=光源からの光が照射していない */
		*id= *is = 0;
	}else{
	  /* 陰影処理 */
	  /* 拡散反射光の強度 */
	  *id = src.kd * dot_N_L * Q.ii / (r * r);
	  /* 鏡面反射光の強度 */
	  R = sub(scale(2 * dot_N_L, N), L);	/* 正反射ベクトル */
	  V = sub(E, P);
	  V = normalize(V);	/* 視点へ向かう単位ベクトルを求める */
	  dot_V_R = dot(R, V);
	  *is = src.ks * pow(dot(R, V), src.n) *  Q.ii / (r * r);
	}
}
